
acceptUnmatched True
blocksize 1
bufsize 1048576
follow_symlinks False
force_polling False
inflight None
part_ext Part
partflg 1
post_baseDir None
permCopy True
timeCopy True
randomize False
rename None
post_on_start False
sleep 5
nodupe_ttl 0

callback gather.file
callback post.message
